// only .text section is copied to bootloader so static strings do not work
// TODO: use SATA driver

#include <lib/elf.h>
#include <lib/memory_layout.h>
#include <lib/type.h>

#include <driver/ide.h>


#define SCRATCH 0x10000


// segment can consist of many sectors and its address must be rounded to sector boundary
static void
read_segment(uint64_t const offset, uint64_t const count, uint64_t const dst)
{
        enum {sector_size = 512};

        uint64_t const aligned_dst = dst - offset%sector_size;

        // ignore first 8 sectors (bootloader)
        for (uint32_t sector = (offset/sector_size) + 8, i = 0; i < count; ++sector, i += sector_size)
                ide_read_sector(sector, aligned_dst + i);
}




void
boot_main(void)
{
        // TODO: strict aliasing?
        struct elf_file_header *const elf = (struct elf_file_header *)SCRATCH; // free for use (same as xv6)

        // TODO: hope that program header fits in
        read_segment(0, 4096, (uint64_t)elf);

        // TODO: refactor
        if (elf->ident[0] != '\x7F' || elf->ident[1] != 'E' || elf->ident[2] != 'L' || elf->ident[3] != 'F')
                return;

        // must be 64-bit
        if (elf->ident[4] != 2 || elf->machine != 0x3E)
                return;


        // TODO: strict aliasing?
        // TODO: set BSS to zero
        // read program segments for each entry in program header table
        struct elf_program_header *ph = (struct elf_program_header *)(SCRATCH + elf->phoff);
        for (uint16_t i = 0; i < elf->phnum; ++i)
                read_segment(ph[i].offset, ph[i].filesz, ph[i].vaddr);

        // call kernel code
        ((void (*)(void))elf->entry)();
}
