# reads given sector from disk
# simpler in assembly as we do not use inline assembly
# TODO: use small assembly functions (such as outb) and call them from C
# reference: http://wiki.osdev.org/ATA_PIO_Mode

#define PRIMARY 0x1F0

#define RDY 0x40
#define BSY 0x80

# TODO: read multiple sector? for LBA28 we can read at most 255 sectors (is it enough?)
# first argument is the sector, the second is pointer to 512 bytes where to load the sector
.global ide_read_sector
ide_read_sector:
        call    wait

        # one sector
        movq    $(PRIMARY + 2), %rdx
        movq    $1, %rax
        outb    %al, (%dx)

        # sector low
        movq    $(PRIMARY + 3), %rdx
        movq    %rdi, %rax
        outb    %al, (%dx)

        # sector mid
        movq    $(PRIMARY + 4), %rdx
        shrq    $0x8, %rax
        outb    %al, (%dx)

        # sector high
        movq    $(PRIMARY + 5), %rdx
        shrq    $0x8, %rax
        outb    %al, (%dx)

        # master 0xE0 and sector highest 4 bits
        movq    $(PRIMARY + 6), %rdx
        shrq    $0x8, %rax
        andq    $0xF, %rax
        orq     $0xE0, %rax
        outb    %al, (%dx)

        # read sector
        movq    $(PRIMARY + 7), %rdx
        movq    $0x20, %rax
        outb    %al, (%dx)

        # wait and read result
        call wait
        movq    $PRIMARY, %rdx
        movq    $0x80, %rcx # 128 times copy 4 bytes
        movq    %rsi, %rdi
        cld
        rep insl (%dx), %es:(%rdi)

        ret


wait:
        # TODO: we move each iteration (unnecessary)
        movq    $(PRIMARY + 7), %rdx
        inb     (%dx), %al

        # wait until drive is not busy
        testb   $BSY, %al
        jnz     wait

        # wait until drive is ready
        # TODO: we can combine busy and ready but this way it seems more readable
        testb   $RDY, %al
        jz      wait

        ret
