#include <lib/memory_layout.h>
#include <lib/mmu.h>

.global load_cr3
load_cr3:
        movq %rdi, %cr3
        ret


.global load_gdt
load_gdt:
        lgdt    (%rdi)
        ret


.global load_idt
load_idt:
        lidt    (%rdi)
        ret

.global load_tr
load_tr:
        movq    %rdi, %rax
        ltr     %ax
        ret


.global sfence
sfence:
        sfence
        ret


.global switch_stack
switch_stack:
        movq    %rdi, %rsp
        movabsq $main, %rax
        jmpq    *%rax


.global load_msr
load_msr:
        movq    %rdi, %rcx

        # split 64-bit input into two 32-bit values
        movq    %rsi, %rax
        shr     $32, %rsi
        movq    %rsi, %rdx

        wrmsr
        ret


.global store_msr
store_msr:
        movq    %rdi, %rcx
        rdmsr
        shl     $32, %rdx
        orq     %rdx, %rax
        ret

.global store_cr2
store_cr2:
        movq    %cr2, %rax
        retq

.global swapgs
swapgs:
        swapgs
        ret


.global core_address
core_address:
        rdgsbase %rax
        ret


# TODO: maintain C calling conventions
.global syscall_return
syscall_return:
        movq    %rdi, %rcx # TODO: restore rip
        movq    %rsi, %rsp

        movq    $RFLAGS_INTERRUPT_FLAG, %r11 # TODO: restore rflags
        swapgs

        # zero out caller saved registers to avoid leaking information (preserse RAX - return value, RCX - RIP, RSP, R11 - RFLAGS)
        # TODO: restore r11 with RFLAGS
        # TODO: how to optimize or avoid
        movq    $0, %rdx
        movq    $0, %rdi
        movq    $0, %rsi
        movq    $0, %r8
        movq    $0, %r9
        movq    $0, %r10
        sysretq


# non-preemptible
# TODO: disable interrupts
.global syscall_call
syscall_call:
        swapgs
        movq    %rsp, %gs:8 # save stack pointer
        movq    %gs:0, %rsp # switch to kernel stack

        # check if system call is valid
        # TODO: write test for this check
        cmpq    $7, %rax # TODO: macro for number of syscalls
        jb      valid # unsigned
        movq    $-1, %rax
        movq    %gs:8, %rsp # restore user stack pointer
        swapgs
        sysretq

valid:
        pushq   %gs:8
        pushq   %rcx # save instruction pointer

        # restore calling convention
        movq    %r10, %rcx

        movabsq $syscalls, %r10
        movq    (%r10,%rax,8), %rax
        callq   *%rax

        popq    %rcx # restore instruction pointer
        popq    %rsp
        swapgs

        # zero out caller saved registers to avoid leaking information (preserse RAX - return value, RCX - RIP, RSP, R11 - RFLAGS)
        # TODO: restore r11 with RFLAGS
        # TODO: how to optimize or avoid
        movq    $0, %rdx
        movq    $0, %rdi
        movq    $0, %rsi
        movq    $0, %r8
        movq    $0, %r9
        movq    $0, %r10

        sysretq


.global store_cr0
store_cr0:
        movq %cr0, %rax
        ret
.global store_cr4
store_cr4:
        movq %cr4, %rax
        ret


# TODO: this needs to be all in assembly so we have precise control over registers (thus avoid pushing too many)
# 1. store RIP into `info` (should be simple if address is fixed)
# 2. change RIP to user mode handler RIP
.global timer_interrupt
timer_interrupt:
        swapgs
        pushq   %rax
        pushq   %rcx
        pushq   %rdx

        # increase penalty
        movabsq KERNEL_INFO_ADDR_PENALTY, %rax
        incq    %rax

        # kill container (penalty == 3)
        cmpq    $3, %rax
        jne     notify

        callq   kill

notify:
        movabsq %rax, KERNEL_INFO_ADDR_PENALTY

        # copy RIP to info
        movq    24(%rsp), %rax
        movabsq %rax, KERNEL_INFO_ADDR

        # change RIP in the interrupt stack
        movq    %gs:16, %rax
        movq    %rax, 24(%rsp)

        # acknowledge interrupt timer
        # TODO: we could use load_msr, but it would be more error prone
        movq    $MSR_APIC_EOI, %rcx
        xorq    %rdx, %rdx
        movq    %rdx, %rax
        wrmsr

        # set new deadline for timer (in case the container does not call yield)
        callq   store_tsc
        movq    %rax, %rcx
        # 10 ms (10*tsc_rate_ms)
        movq    $10, %rax
        mulq    %gs:24
        addq    %rcx, %rax

        movq    $MSR_APIC_TSC_DEADLINE, %rcx
        movq    %rax, %rdx
        shrq    $32, %rdx
        wrmsr

        popq    %rdx
        popq    %rcx
        popq    %rax
        swapgs
        iretq


.global abort_gp_asm
abort_gp_asm:
        movq    (%rsp), %rdi
        callq   abort_gp
