#ifndef KERNEL_AMD64_H
#define KERNEL_AMD64_H

#include <lib/type.h>


struct tss {
        uint32_t reserved0;
        uint32_t rsp0[2];
        uint32_t rsp1[2];
        uint32_t rsp2[2];
        uint32_t reserved1[2];
        uint32_t ist1[2];
        uint32_t ist2[2];
        uint32_t ist3[2];
        uint32_t ist4[2];
        uint32_t ist5[2];
        uint32_t ist6[2];
        uint32_t ist7[2];
        uint32_t reserved2[2];
        uint16_t reserved3;
        uint16_t io_map_base_address;
};


struct gdt {
        uint64_t entry[6];
};

struct gdtr {
        uint16_t limit;
        uint16_t base_address[4];
};


// interrupt and trap gate (long mode has no call gate support)
struct gate_descriptor {
        uint16_t offset0;
        uint16_t selector;
        uint8_t ist;
        uint8_t properties;
        uint16_t offset1;
        uint32_t offset2;
        uint32_t reserved0;
};


struct idt {
        struct gate_descriptor gates[256];
};



void
load_cr3(uint64_t pml4_address);

void
load_gdt(uint64_t gdtr_address);

void
load_idt(uint64_t idtr_address);

void
load_tr(uint64_t selector);

void
sfence(void);

void
switch_stack(uint64_t stack_address);

void
load_msr(uint64_t msr_address, uint64_t value);

uint64_t
store_msr(uint64_t msr_address);

uint64_t
store_cr2(void);


void
swapgs(void);

uint64_t
core_address(void);

void
syscall_return(uint64_t rip, uint64_t rsp);

// TODO: remove as it should never be called directly
void
syscall_call(void);

#endif
