#include <lib/memory_layout.h>
#include <lib/mmu.h>

# segment selectors, first 3 bits are used for flags
.set KERNEL_CODE_SEGMENT_SELECTOR, 0x8 # second entry


#define LOW(address) ((address) - ap_entry_start + 0x5000ul)

.global ap_entry_start
ap_entry_start:
        .code16

        xorw    %ax, %ax
        movw    %ax, %ds
        movw    %ax, %es
        movw    %ax, %ss

        # enable PAE and global pages
        movw    $(CR4_PAE | CR4_PGE), %ax
        movl    %eax, %cr4


        # load page table
        movl    (0x5000 - 8), %eax
        movl    %eax, %cr3

        # enable long mode
        movl    $EFER, %ecx
        rdmsr
        orl     $EFER_LME, %eax
        wrmsr

        # enable paging and protection (as we are going from real mode to long mode)
        movl    %cr0, %eax
        orl     $(CR0_PG | CR0_WP | CR0_PE), %eax
        movl    %eax, %cr0

        # load Global Descriptor Table
        lgdt    LOW(ap_gdt_descriptor)

        ljmpl   $KERNEL_CODE_SEGMENT_SELECTOR, $(LOW(long_mode))


# global descriptor table (GDT)
# PAGE TABLE ENTRY MUST BE WRITABLE
.align 8
gdt:
        SEGMENT_NULL
        SEGMENT_KERNEL_CODE

ap_gdt_descriptor:
        .word   0x1F // sizeof gdt - 1
        .quad   LOW(gdt)



long_mode:
        .code64

        xorq    %rax, %rax
        movq    %rax, %ds
        movq    %rax, %es
        movq    %rax, %fs
        movq    %rax, %gs
        movq    %rax, %ss

        # TODO: what is scope of this MSR? (without it we get reboot)
        # enable no-execute page support
        movq    $0xC0000080, %rcx
        rdmsr
        orq     $(EFER_NXE | EFER_SCE), %rax
        wrmsr

        movq    %cr4, %rax
        orq     $(CR4_SMEP | CR4_FSGSBASE), %rax # TODO: my cpu does not support SMAP
        movq    %rax, %cr4

        movq    %cr0, %rax
        andq    $(~CR0_EM), %rax
        orq     $2, %rax
        movq    %rax, %cr0

        movq    %cr4, %rax
        orq     $(CR4_SMEP | CR4_OSXSAVE | CR4_OSXMMEXCPT | CR4_OSFXSR | CR4_FSGSBASE), %rax # TODO: my cpu does not support SMAP
        movq    %rax, %cr4

        # enable AVX
        movq    $0, %rcx
        xgetbv
        orq     $(XCR0_YMM | XCR0_SSE), %rax
        xsetbv

        # disable alignment check (necessary for SMAP)
        #clac

        # point stack pointer to core's stack
        movq    (0x5000 - 16), %rsp

        # call to C code
        movq    (0x5000 - 24), %rdi
        movabsq $ap_main, %rax
        call    *%rax

spin:
        jmp spin


.global ap_entry_end
ap_entry_end:
