# only purpose is check CPU support and setup minimum required bits and point stack pointer to top of our data segment (startup stack)

#include <lib/memory_layout.h>
#include <lib/mmu.h>

.global _start
_start:
        # disable master (0x20) and slave (0xA0) PIC
        movq    $0xFF, %rax
        outb    %al, $0x21
        outb    %al, $0xA1


        # CHECKS
        # test if CPU supports required features
        movq    $0x80000001, %rax
        cpuid

        testq   $(CPUID_LM | CPUID_NX | CPUID_PGE | CPUID_SYSCALLSYSRET | CPUID_MSR | CPUID_TSC), %rdx
        jz      _start


        # extended features
        movq    $0x7, %rax
        movq    $0, %rcx
        cpuid

        testq   $(CPUID_SMAP | CPUID_SMEP | CPUID_FSGSBASE), %rbx
        jz      _start


        movq    $0x1, %rax
        cpuid
        testq   $(CPUID_RDRAND | CPUID_AVX | CPUID_TSC_DEADLINE | CPUID_X2APIC), %rcx
        jz      _start


        # CPUID: Advanced Power Management Information
        movq    $0x80000007, %rax
        cpuid
        testq   $CPUID_TSC_INVARIANT, %rdx
        jz      _start


        # TODO: check AVX2 support


        # enable no-execute page support and system call extensions
        movq    $EFER, %rcx
        rdmsr
        orq     $(EFER_NXE | EFER_SCE), %rax
        wrmsr

        movq    %cr0, %rax
        andq    $(~CR0_EM), %rax
        orq     $2, %rax
        movq    %rax, %cr0

        movq    %cr4, %rax
        orq     $(CR4_SMEP | CR4_OSXSAVE | CR4_OSXMMEXCPT | CR4_OSFXSR | CR4_FSGSBASE), %rax # TODO: my cpu does not support SMAP
        movq    %rax, %cr4

        # enable AVX
        movq    $0, %rcx
        xgetbv
        orq     $(XCR0_YMM | XCR0_SSE), %rax
        xsetbv

        # disable alignment check (necessary for SMAP)
        #clac

        # stack and jump
        movabsq $(KERNEL_BASE + 0x1000000), %rsp # TODO: do not hardcode, instead make .data segment large

        movabsq $bootstrap, %rax
        jmpq *%rax
