#ifndef CAPABILITY_H
#define CAPABILITY_H

#include "sha256.h"
#include "type.h"


struct capability {
        uint64_t perm;
        uint64_t hash;
};


static inline bool
capability_init(struct capability *const c, uint64_t const perm, uint64_t const key)
{
        uint32_t const message[4] = {perm >> 32, perm, key >> 32, key};

        uint32_t hash[8];
        if (!sha256(4, message, hash))
                return false;

        *c = (struct capability){
                .perm = perm,
                .hash = (((uint64_t)hash[0]) << 32) | hash[1],
        };
        return true;
}

static inline bool
capability_check(struct capability const c, uint64_t const key)
{
        uint32_t const message[4] = {c.perm >> 32, c.perm, key >> 32, key};

        uint32_t hash[8];
        if (!sha256(4, message, hash))
                return false;

        return (hash[0] == (uint32_t)(c.hash >> 32)) && (hash[1] == (uint32_t)c.hash);
}


#endif
