#ifndef MEMORY_LAYOUT_H
#define MEMORY_LAYOUT_H

/*
TODO: USER_PAGES does need less storage than 512 GiB
TODO: can stack pointer become negative? (would need extra guard page)

                                                 Write permisions:
                                                 kernel/user
256 TiB ->      +-------------------------------+
                | heap                          |
kernel_end  ->  +-------------------------------+
                | kernel                        | W/-
KERNEL_BASE ->  +-------------------------------+       0xFFFFFF8000000000 (511 pml4 entry)
                | kernel stack (core 0)         |
                +-------------------------------+
                | guard page                    |
                +-------------------------------+
                | kernel stack (core 1)         |
                +-------------------------------+
                | guard page                    |
                +-------------------------------+
                | kernel stack (core n)         |
                +-------------------------------+
                | guard page                    |
                +-------------------------------+       0xFFFFFF0000000000 (510 pml4 entry)
                | container info (per core)     |
                +-------------------------------+ W/-   0xFFFFFE8000000000 (509 pml4 entry)
                | container info (per core)     |
                +-------------------------------+ -/R   0xFFFFFE0000000000 (508 pml4 entry)

                ^ ----- ACCESSIBLE        ----- ^

                           HOLE (AMD64)

                V -----  ACCESSIBLE       ----- V       0x7FFFFFFFFFFFFFFF

                | recursively mapped PML4       | R/R
USER_PML4 ->    +-------------------------------+       0x0000700000000000
                |                               | R/R   TODO: needs less memory
USER_PAGES ->   +-------------------------------+       0xFE0000000000
                |                               |
0 ->            +-------------------------------+
*/


// at 512 GB virtual address (second entry in PML4)
// used for simplicity to share PDP and PD in pagetable with physical mapping at boot
// must be multiple of 512 GB (i.e., first 39 bits must be zero) 
#define KERNEL_BASE 0xFFFFFF8000000000ul // canonical address form
#define EXTENDED_MEMORY 0x100000ul
#define PHYSICAL_TOP 0x10000000ul // 256 MiB


#define V2P(virtual_address) ((virtual_address) - KERNEL_BASE)



#define USER_TOP (512*2*1024*1024ul) // TODO:

#define INFO_ADDR 0xFFFFFE0000000000ul

#define KERNEL_INFO_ADDR 0xFFFFFE8000000000ul
#define KERNEL_INFO_ADDR_PENALTY 0xFFFFFE8000000008ul


#endif
