#ifndef MMU_H
#define MMU_H


#define SEGMENT_NULL \
        .word 0, 0; \
        .byte 0, 0, 0, 0

#define SEGMENT_KERNEL_CODE \
        .word 0, 0; \
        .byte 0, 0x98 /* present bit and required 1 bit */, 0x20 /* long mode */, 0

#define SEGMENT_USER_CODE \
        .word 0, 0; \
        .byte 0, 0xF8 /* present bit and required 1 bit */, 0x20 /* long mode */, 0

#define SEGMENT_TSS \


#define PAGE_SIZE       4096



// Control Register 0
#define CR0_PE  0x1             // Protected Mode Enable
#define CR0_EM  0x4             // Emulation
#define CR0_WP  0x10000         // Write Protect (kernel can't write to read-only pages)
#define CR0_PG  0x80000000      // Paging


// Control Register 4
#define CR4_PAE         0x20     // Physical Address Extension
#define CR4_PGE         0x80     // Page Global Enabled (TODO)
#define CR4_PCE         0x100    // Performance-Monitoring Counter
#define CR4_OSFXSR      0x200    // Operating system support for FXSAVE and FXRSTOR instructions
#define CR4_OSXMMEXCPT  0x400    // Operating system support for unmasked SIMD floating-point exceptions
#define CR4_FSGSBASE    0x10000  // Enable RDFSBASE, RDGSBASE, WRFSBASE, and WRGSBASE instructions
#define CR4_OSXSAVE     0x40000  // XSAVE and Processor Extended States
#define CR4_SMEP        0x100000 // Supervisor Mode Execution Prevention
#define CR4_SMAP        0x200000 // Supervisor Mode Access Protection


// Extended Control Register 0
#define XCR0_SSE 0x2 // SSE (must be enabled for AVX support)
#define XCR0_YMM 0x4 // AVX


// Extended Feature Enable Register
#define EFER 0xC0000080ul
#define EFER_SCE 0x1   // System Call Extensions
#define EFER_LME 0x100 // Long Mode Enable
#define EFER_NXE 0x800 // No-Execute Enable


// Page Table Entry
#define PTE_P   0x1ul     // Present
#define PTE_W   0x2ul     // Write
#define PTE_U   0x4ul     // User
#define PTE_PS  0x80ul    // Page Size
#define PTE_G   0x100ul   // Global Page
#define PTE_NX  0x8000000000000000ul // No Execute


// Model-Specific Registers
#define MSR_STAR 0xC0000081ul // sysret and syscall selectors
#define MSR_LSTAR 0xC0000082ul // syscall target %rip
#define MSR_SFMASK 0xC0000084ul // system call rFLAGS mask
#define MSR_KERNEL_GS_BASE 0xC0000102ul


// CPUID EAX=1 feature bits
#define CPUID_X2APIC       0x200000ul
#define CPUID_TSC_DEADLINE 0x1000000ul
#define CPUID_AVX          0x10000000ul
#define CPUID_RDRAND       0x40000000ul // random number generator


// CPUID Fn0000_0001_EDX  Feature Identifiers
#define CPUID_TSC 0x10ul
#define CPUID_MSR 0x20ul
#define CPUID_APIC 0x20ul
#define CPUID_SYSCALLSYSRET 0x800ul
#define CPUID_PGE 0x2000ul
#define CPUID_NX 0x100000ul
#define CPUID_LM 0x20000000ul

// EAX=7, ECX=0: Extended Features
#define CPUID_FSGSBASE 0x0ul
#define CPUID_SMEP 0x80ul
#define CPUID_SMAP 0x100000ul

// Power Management
#define CPUID_TSC_INVARIANT 0x100ul


#define RFLAGS_INTERRUPT_FLAG 0x200ul


// x2APIC
#define MSR_APIC_EOI 0x80Bul
#define MSR_APIC_TSC_DEADLINE 0x6E0ul


#endif
