#include <lib/memory_layout.h>

.global gen_ud_fault
gen_ud_fault:
        ud2
        retq


.global gen_gp_fault
gen_gp_fault:
        movq    %cr2, %rax
        retq

.global gen_int
gen_int:
        int     $0x0
        retq


.global fault_de
fault_de:
        # save registers
        pushq   %rax
        pushq   %rbx
        pushq   %rcx
        pushq   %rdx
        pushq   %rbp
        pushq   %rsi
        pushq   %rdi
        pushq   %r8
        pushq   %r9
        pushq   %r10
        pushq   %r11
        pushq   %r12
        pushq   %r13
        pushq   %r14
        pushq   %r15

        movq    152(%rsp), %rdi
        movq    144(%rsp), %rsi
        movq    136(%rsp), %rdx
        movq    128(%rsp), %rcx
        movq    120(%rsp), %r8
        movabsq $de_fault, %rax
        callq   *%rax

        popq    %r15
        popq    %r14
        popq    %r13
        popq    %r12
        popq    %r11
        popq    %r10
        popq    %r9
        popq    %r8
        popq    %rdi
        popq    %rsi
        popq    %rbp
        popq    %rdx
        popq    %rcx
        popq    %rbx
        popq    %rax

        retq    $32 


# TODO: somehow reuse the register push/pop (we could have one handler like xv6)
.global fault_xf
fault_xf:
        # save registers
        pushq   %rax
        pushq   %rbx
        pushq   %rcx
        pushq   %rdx
        pushq   %rbp
        pushq   %rsi
        pushq   %rdi
        pushq   %r8
        pushq   %r9
        pushq   %r10
        pushq   %r11
        pushq   %r12
        pushq   %r13
        pushq   %r14
        pushq   %r15

        movq    152(%rsp), %rdi
        movq    144(%rsp), %rsi
        movq    136(%rsp), %rdx
        movq    128(%rsp), %rcx
        movq    120(%rsp), %r8
        movabsq $xf_fault, %rax
        callq   *%rax

        popq    %r15
        popq    %r14
        popq    %r13
        popq    %r12
        popq    %r11
        popq    %r10
        popq    %r9
        popq    %r8
        popq    %rdi
        popq    %rsi
        popq    %rbp
        popq    %rdx
        popq    %rcx
        popq    %rbx
        popq    %rax

        retq    $32 # can't use iretq as that would load CS from GDT in kernel space -> page fault


.global fault_pf
fault_pf:
        # save registers
        pushq   %rax
        pushq   %rbx
        pushq   %rcx
        pushq   %rdx
        pushq   %rbp
        pushq   %rsi
        pushq   %rdi
        pushq   %r8
        pushq   %r9
        pushq   %r10
        pushq   %r11
        pushq   %r12
        pushq   %r13
        pushq   %r14
        pushq   %r15

        movq    160(%rsp), %rdi
        movq    152(%rsp), %rsi
        movq    144(%rsp), %rdx
        movq    136(%rsp), %rcx
        movq    128(%rsp), %r8
        movq    120(%rsp), %r9
        movabsq $pf_fault, %rax
        callq   *%rax

        popq    %r15
        popq    %r14
        popq    %r13
        popq    %r12
        popq    %r11
        popq    %r10
        popq    %r9
        popq    %r8
        popq    %rdi
        popq    %rsi
        popq    %rbp
        popq    %rdx
        popq    %rcx
        popq    %rbx
        popq    %rax

        retq    $40


# TODO: kernel does not push any stack so if we receive interrupt before pushing RIP user will hang (and eventually be killed)
# TODO: we could make kernel to push RIP or have separate read-only stack for handler
# TODO: save only what is necessary
.global timer_interrupt_asm
timer_interrupt_asm:
        # push RIP from before the interrupt onto the stack so after calling handler the
        # execution is returned to address from before the interrupt
        # TODO: not atomic, so if interrupted the user gets screwed
        # TODO: we could have read only stack for interrupts
        movq    %rax, -16(%rsp) # backup RAX
        movabsq INFO_ADDR, %rax
        pushq   %rax
        movq    -8(%rsp), %rax # restore RAX

        # TODO: do not save registers if next container is us
        pushq   %rax
        pushq   %rbx
        pushq   %rcx
        pushq   %rdx
        pushq   %rbp
        pushq   %rsi
        pushq   %rdi
        pushq   %r8
        pushq   %r9
        pushq   %r10
        pushq   %r11
        pushq   %r12
        pushq   %r13
        pushq   %r14
        pushq   %r15

        # TODO: oxsave? aligned? (is it enough to store only registers? is there some other state?)
        # save AVX registers
        vmovdqu %ymm0, -32(%rsp)
        vmovdqu %ymm1, -64(%rsp)
        vmovdqu %ymm2, -96(%rsp)
        vmovdqu %ymm3, -128(%rsp)
        vmovdqu %ymm4, -160(%rsp)
        vmovdqu %ymm5, -192(%rsp)
        vmovdqu %ymm6, -224(%rsp)
        vmovdqu %ymm7, -256(%rsp)
        vmovdqu %ymm8, -288(%rsp)
        vmovdqu %ymm9, -320(%rsp)
        vmovdqu %ymm10, -352(%rsp)
        vmovdqu %ymm11, -384(%rsp)
        vmovdqu %ymm12, -416(%rsp)
        vmovdqu %ymm13, -448(%rsp)
        vmovdqu %ymm14, -480(%rsp)
        vmovdqu %ymm15, -512(%rsp)
        subq    $512, %rsp

        callq   timer_interrupt

        addq    $512, %rsp
        vmovdqu -32(%rsp), %ymm0
        vmovdqu -64(%rsp), %ymm1
        vmovdqu -96(%rsp), %ymm2
        vmovdqu -128(%rsp), %ymm3
        vmovdqu -160(%rsp), %ymm4
        vmovdqu -192(%rsp), %ymm5
        vmovdqu -224(%rsp), %ymm6
        vmovdqu -256(%rsp), %ymm7
        vmovdqu -288(%rsp), %ymm8
        vmovdqu -320(%rsp), %ymm9
        vmovdqu -352(%rsp), %ymm10
        vmovdqu -384(%rsp), %ymm11
        vmovdqu -416(%rsp), %ymm12
        vmovdqu -448(%rsp), %ymm13
        vmovdqu -480(%rsp), %ymm14
        vmovdqu -512(%rsp), %ymm15

        popq    %r15
        popq    %r14
        popq    %r13
        popq    %r12
        popq    %r11
        popq    %r10
        popq    %r9
        popq    %r8
        popq    %rdi
        popq    %rsi
        popq    %rbp
        popq    %rdx
        popq    %rcx
        popq    %rbx
        popq    %rax

        retq
